
# This file was generated by 'versioneer.py' (0.18) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.
from __future__ import absolute_import
import json

version_json = '''
{
 "date": "2021-11-25T15:11:44-0500",
 "dirty": false,
 "error": null,
 "full-revisionid": "3584072c4a38f8ea2d7e08fdb6db177aa82f0bbc",
 "version": "1.14.0"
}
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)
